## Data Model component

In this tutorial, we will learn how to use data model based UI. 

### Problems

We often encounter following two problems during development:

* Share same variable across multiple components

When we have multiple components that share the same variable, we need to pass the variable to each component and send event to frontend. This is not only cumbersome, but also error-prone.

* Render large list/table 

Currently we need to create a huge list of UIs for each data item, which is hard to maintain and slow to render.


### `mui.DataModel` and `bind_fields`

#### Overview

`mui.DataModel` is a data model component container, it wraps a `dataclasses` object. To let child components access value of your `dataclasses`, use `bind_fields` to bind a value to a child component.

```Python

from tensorpc.dock import mui, marker
import dataclasses

@dataclasses.dataclass
class Model:
    name: str 
    count: int 

class App:
    @marker.mark_create_layout
    def layout(self):
        model = Model("test", 1)
        # this draft is only used for binding fields, don't use it to update values
        model_draft = mui.DataModel.get_draft_external(model)
        self.dm = mui.DataModel(Model("test", 1), [
            mui.Markdown().bind_fields(value=model_draft.name),
            mui.Markdown().bind_fields(value=f"to_string({model_draft.count})"),
            mui.Button("increment", self._handle_button),
            mui.Button("increment With Draft Update", self._handle_button_draft),
        ])
        return mui.VBox([
            self.dm
        ])

    async def _handle_button(self):
        self.dm.model.count += 1
        # explicitly sync whole model to update UI, slow for large model.
        await self.dm.sync_model()

    async def _handle_button_draft(self):
        # use draft to update data, inspired by `immer`.
        # you can't update data directly with plain dataclasses object, you must
        # use draft.
        draft = self.dm.get_draft()
        draft.count += 1
```

#### JMESPath

We use [jmespath](https://jmespath.org/) as data query language for `bind_fields`, you can use `jmespath` expression to query data from `mui.DataModel`.

**WARNING**: if jmespath expression is invalid, it will return `undefined` instead of `null`, also all jmespath query results will be convert to `undefined` even if it's a `null` (None) set by user because almost all component props accept `undefined` instead of `null`.

* Extended JMESPath

We provide some extended functions for `jmespath`. see `tensorpc.core.datamodel.funcs` for more details.

#### Draft and JMESPath

We can see we use draft object instead of string as `jmespath`, this is a grammar sugar to make attribute error check works with modern ide. when you create a draft and access fields, a jmespath expression is generated automatically.

**WARNING**: The most important difference between draft and jmespath is that evaluation result of draft expr will be `null` if eval failed in both frontend and backend (`evaluate_draft_ast_noexcept`). e.g. `draft.a.b` is still a valid expr of `model.a` is `null`. This feature is very important for frontend since we can't check and handle the error in draft expr via `if`.

#### Update UI by change Model

since all child components are binded to model, you can update these binded fields to directly update all related components.

`_handle_button` method sync whole model to frontend, so related `Markdown` components will be updated.

Update whole model is simple, but slow for large model. 

#### Update UI by draft

We recommend to use `draft` to update model, it's inspired by `immer`. all changes you made to draft will be recorded and applied in frontend via a `JSONPath` like way, so it's very fast and simple. To make life easier, we record all draft operation inside any event handler, you don't need to call any async function to send events to frontend.

To use draft update, you need to get draft object from `mui.DataModel`.

Keep in mind that draft operations is recorded and sent **AFTER** event handler run. If you want to perform model update and send event immediately, you can use `draft_update`:

```py
async with self.dm.draft_update() as draft:
    draft.count += 1
```

Events will be sent immediately after `async with` block.

### Optional UI

Recall the jmespath invalid behavior, all binded fields will become `undefined`. We won't unmount components if their binded prop query result is `undefined`. To make components invisible instead of render with `undefined` props, we can use `mui.MatchCase`.

```Python

from tensorpc.dock import mui, marker
import dataclasses
from typing import Optional

@dataclasses.dataclass
class ModelItem:
    name: str 

@dataclasses.dataclass
class Model:
    item: Optional[ModelItem]

class App:
    @marker.mark_create_layout
    def layout(self):
        model = Model(ModelItem("## test"))
        # this draft is only used for binding fields, don't use it to update values
        model_draft = mui.DataModel.get_draft_external(model)
        self.dm = mui.DataModel(model, [
            mui.Switch("Show", self._handle_sw, init_value=True),
            mui.MatchCase.binary_selection(True, 
                mui.Markdown().bind_fields(value="item.name"), 
                mui.Markdown("## Hidden")
            ).bind_fields(condition=f"{model_draft.item} != null"),
        ])
        return mui.VBox([
            self.dm
        ])

    async def _handle_sw(self, enable):
        draft = self.dm.get_draft()
        draft.item = None if not enable else ModelItem("## test")
```

### Nested Data Model

Currently nested data model has limited support. All components can only see closest parent model, so you can't access grandparent model directly. We will improve this in future.

### Nested Query

Assume we have a data list, we want to put a list left and a detail panel on right to show detail of selected list item. 

#### Definition

* `Source`

Data model come from `DataModel`, data list components such as `DataFlexBox`, `SimpleFlow` and resource loaders in `three` can be treated as **Source**. child components can use `bind_fields` to query data from source.

* `Relay`

When you use `DataSubQuery` or data list components with `dataList` prop binded to other source, they are treated as **Relay**. 

For each child components, we construce a **query chain**:

```

Source -> Relay -> Relay -> ... -> Relay -> consumer

```

#### All data model components

* `DataModel`: source comes from `dataObject` prop. this component can't bind any field from other source.

* `DataSubQuery`: this component can't be used as source, it can only comsume data from other source. 

* data list components: these components can be used as source when their `dataList` prop isn't binded to other source. you can also bind `dataList` to other source.

* `DataPortal`: can't be used as `Relay`. components set in `DataPortal` will be treated as source even if it's a `Relay`.

* resource loaders: 3D assets loader in `three` can be used as source, you can't bind fields to other source.

* `Flow`: can only be used as source, you can't bind fields to other source. 

#### Examples

```Python

from tensorpc.dock import mui, marker
import dataclasses
from typing import Optional

@dataclasses.dataclass
class ModelItem:
    name: str 
    count: int

@dataclasses.dataclass
class Model:
    items: list[ModelItem]
    cur: Optional[int] = None

class App:
    @marker.mark_create_layout
    def layout(self):
        model = Model([
            ModelItem("item1", 1),
            ModelItem("item2", 2),
            ModelItem("item3", 3),
        ])
        # this draft is only used for binding fields, don't use it to update values
        model_draft = mui.DataModel.get_draft_external(model)
        btn = mui.Button("").bind_fields(name="name")
        btn.event_click.on_standard(self._handle_list_item)
        self.dm = mui.DataModel(model, [
            mui.HBox([
                mui.DataFlexBox(btn).prop(flex=1, flexFlow="column", height="100%", border="1px solid red", idKey="name").bind_fields(dataList="items"),
                mui.VBox([
                    mui.MatchCase([
                        mui.MatchCase.Case(True, mui.HBox([
                            mui.Markdown().bind_fields(value="getitem(items, cur).name"),
                            mui.Markdown().bind_fields(value="to_string(getitem(items, cur).count)"),
                        ])),
                        mui.MatchCase.Case(False, mui.Markdown("## Nothing")),
                    ]).bind_fields(condition=f"{model_draft.cur} != null"),
                ]).prop(flex=2)
            ])
        ])
        return mui.VBox([
            self.dm
        ]).prop(height="100%", width="100%")

    def _handle_list_item(self, event: mui.Event):
        index = event.indexes[0]
        draft = self.dm.get_draft()
        draft.cur = index

```

We can see that we have to declare `getitem` in each components in detail panel. To resolve this, we need to use `DataSubQuery` to forward some childs in main data model:

```Python
from tensorpc.dock import mui, marker
import dataclasses
from typing import Optional

@dataclasses.dataclass
class ModelItem:
    name: str 
    count: int

@dataclasses.dataclass
class Model:
    items: list[ModelItem]
    cur: Optional[int] = None

class App:
    @marker.mark_create_layout
    def layout(self):
        model = Model([
            ModelItem("item1", 1),
            ModelItem("item2", 2),
            ModelItem("item3", 3),
        ])
        # this draft is only used for binding fields, don't use it to update values
        model_draft = mui.DataModel.get_draft_external(model)
        btn = mui.Button("").bind_fields(name="name")
        btn.event_click.on_standard(self._handle_list_item)
        self.dm = mui.DataModel(model, [
            mui.HBox([
                mui.DataFlexBox(btn).prop(flex=1, flexFlow="column", height="100%", border="1px solid red", idKey="name").bind_fields(dataList="items"),
                mui.VBox([
                    mui.Button("Close", self._handle_close),
                    mui.MatchCase.binary_selection(True, mui.DataSubQuery("getitem(items, cur)", [mui.HBox([
                        mui.Markdown().bind_fields(value="name"),
                        mui.Markdown().bind_fields(value="to_string(count)"),
                    ])])).bind_fields(condition=f"{model_draft.cur} != null"),
                ]).prop(flex=2)
            ])
        ])
        return mui.VBox([
            self.dm
        ]).prop(height="100%", width="100%")

    def _handle_close(self):
        draft = self.dm.get_draft()
        draft.cur = None

    def _handle_list_item(self, event: mui.Event):
        index = event.indexes[0]
        draft = self.dm.get_draft()
        draft.cur = index

```

### Automatic Value Change

Some component (input/textfield/slider/switch/checkbox) support automatic value change, you can use `bind_draft_change` to bind a value to a component, when value changed, draft update is performed automatically.

```Python
from tensorpc.dock import mui, marker
import dataclasses
from typing import Optional

@dataclasses.dataclass
class Model:
    text: str = ""

class App:
    @marker.mark_create_layout
    def layout(self):
        model = Model("")
        # this draft is only used for binding fields, don't use it to update values
        inp = mui.Input("Input Text Here").prop(debounce=300)
        self.dm = mui.DataModel(model, [
            mui.VBox([
                inp,
                mui.Typography().bind_fields(value="text").prop(variant="h1"),
                mui.Typography().bind_fields(value="text").prop(variant="h2"),
                mui.Typography().bind_fields(value="text").prop(variant="h3"),
                mui.Typography().bind_fields(value="text").prop(variant="h4"),
                mui.Typography().bind_fields(value="text").prop(variant="h5"),
                mui.Typography().bind_fields(value="text").prop(variant="h6"),
            ])
        ])
        inp.bind_draft_change_uncontrolled(self.dm.get_draft().text)
        return mui.VBox([
            self.dm
        ]).prop(height="100%", width="100%")

```

### Observe draft expression change

We can also do something when some prop of your data model changed. This is useful for uncontrolled component/operations such as `mui.MonacoEditor` or database opeartion.

```Python
from tensorpc.dock import mui, marker
import dataclasses
from typing import Optional
from tensorpc.core.datamodel.events import DraftChangeEvent

@dataclasses.dataclass
class Model:
    text: str = ""

class App:
    @marker.mark_create_layout
    def layout(self):
        model = Model("")
        # this draft is only used for binding fields, don't use it to update values
        inp = mui.Input("Input Text Here").prop(debounce=300)
        self.md = mui.Markdown()
        self.dm = mui.DataModel(model, [
            mui.VBox([
                inp,
            ])
        ])
        draft = self.dm.get_draft_type_only()
        self.dm.install_draft_change_handler(draft.text, self._handle_model_text_change)
        inp.bind_draft_change_uncontrolled(draft.text)
        return mui.VBox([
            self.dm,
            self.md,
        ]).prop(height="100%", width="100%")

    async def _handle_model_text_change(self, draft_ev: DraftChangeEvent):
        new_value = draft_ev.new_value
        if draft_ev.new_value is not None:
            await self.md.write(new_value)

```

**WARNING**: keep in mind that draft change event value may be `None` during mount/unmount, so you need to check if it's `None` before using it.

### Performance

Compare to regular prop-based render, query-based often have additional overhead to parse and query data from a model. When you use nested query such as `DataSubQuery` or use `bind_fields` in `DataFlexBox`, we need to perform multiple times of query on a query chain mentioned above.

When a data source changed:

* all query chains in binded fields will be re-evaluated

* all components with `bind_fields` will be re-rendered

* `JMESPath` query compilation is cached in whole app, so no compile overhead.

* all components without `bind_fields` won't be re-rendered because they are wrapped in `React.memo`.

* generation of query chains is cached, but it will be cleared when you change layout related to data model components.

So we should keep query simple. If you have a large data model, you should do heavy computation in backend and save them to data model.