## Flow MVC

Modern and complex apps often use MVC pattern to write apps. For flow, since the `flowui.Flow` isn't a controlled component, we need to write additional event and model change handlers. We already write some basic model binder for you.

* Why `flowui.Flow` isn't a controlled component?

The major reason is custom layout for every node. if we use `flowui.DataFlow`, it only use a shared custom component like `mui.DataGrid`, so it can be a controlled component. If we use `flowui.Flow`, we can't treat UI component as a data model, so we can't use controlled method (e.g. `bind_fields`).

### Basic Model Design

The simplest node/edge model is reuse `flowui.Node` and `flowui.Edge`.

```py
@dataclasses.dataclass
class BaseNodeModel(flowui.Node):
    pass

@dataclasses.dataclass
class BaseEdgeModel(flowui.Edge):
    pass

@dataclasses.dataclass
class BaseFlowModel:
    nodes: dict[str, BaseNodeModel]
    edges: dict[str, BaseEdgeModel]

```

Here we use `dict` in model instead of `list` in `flowui.Flow` because it's required for custom node to access data stored in model.

since `flowui.Flow` isn't a fully controlled component, we need to write event handlers and model change handlers to connect UI and data model.

#### `flowui.Flow` to `BaseFlowModel`

Firstly we need to handle events from `flowui.Flow` to update data model.

In source code of `BaseFlowModelBinder`, we register five events. Keep in mind that `flowui.Flow` handle these events internally to update its backend props, so we only need to update data model.

* `event_edge_connection`

Handle edge connection. you can get new edges from event data, then you can get new edges from `flowui.Flow` props and use draft update to update data model. you can also compare nodes in model and ui nodes in props.

* `event_node_delete`

compare nodes in model and ui nodes in props, then use draft update to delete nodes in model.

* `event_edge_delete` 

same as `event_node_delete`, but for edges.

* `event_vis_change`

This event is triggered when you move a node or resize a node. the resize event can be triggered by both manual resizer and custom component mount.

* `event_node_logic_change`

Triggered when you change nodes from frontend. Currently only delete node in frontend can trigger this event.

#### `BaseFlowModel` to `flowui.Flow`

We also need to handle model change to update UI. Use draft change event to detect nodes/edges change in model, then update UI.

```py
dm_comp.install_draft_change_handler(draft.nodes, your_handler)
dm_comp.install_draft_change_handler(draft.nodes, your_handler)
```

We provide a basic model and its event binder for you. You can use it to write your own flow app.

```Python
from tensorpc.dock import mui, three, plus, appctx, mark_create_layout, flowui, models
from tensorpc.core import dataclass_dispatch as dataclasses

@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    type: str = ""
    label: str = ""

@dataclasses.dataclass
class BasicFlowModelRoot(models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel]):
    pass 

def simple_to_ui_node(flow: models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel], node_id: str):
    node = flow.nodes[node_id]
    return flowui.Node(id=node.id, type=node.type, data=flowui.NodeData(label=node.label), position=node.position, width=node.width, height=node.height)

class App:
    class Config(plus.AppInMemory.Config):
        is_horizontal = False
        height = "800px"
    @mark_create_layout
    def my_layout(self):
        self.graph = flowui.Flow([], [], [
            flowui.MiniMap(),
            flowui.Controls(),
            flowui.Background()
        ])
        jv = mui.JsonViewer()
        jv.bind_fields(data="$")
        self.dm = mui.DataModel(BasicFlowModelRoot({}, {}), [
            mui.VBox([
                self.graph,
            ]).prop(flex=1),
            mui.VBox([
                mui.Button("Add Nodes", self._add_nodes_once),
                jv,
            ]).prop(flex=1, overflow="auto", fontSize="12px"),
        ])
        binder = models.flow.BaseFlowModelBinder(
            self.graph, self.dm.get_model, 
            self.dm.get_draft(), 
            simple_to_ui_node)
        binder.bind_flow_comp_with_base_model(self.dm)
        self._added = False
        return mui.HBox([
            self.dm,
        ]).prop(width="100%", height="100%", overflow="hidden")
    
    def _add_nodes_once(self):
        if self._added:
            return 
        nodes = [
            NodeModel(id="1", type="input", label="Drag me and see change", position=flowui.XYPosition(0, 0)),
        ]
        draft = self.dm.get_draft()
        for n in nodes:
            draft.nodes[n.id] = n
        self._added = True

```