## Flow MVC with Custom Node

We can access node props in data model in custom node component.

```Python
from tensorpc.dock import mui, three, plus, appctx, mark_create_layout, flowui, models
from tensorpc.core import dataclass_dispatch as dataclasses
from tensorpc.core.datamodel.draft import create_literal_draft
from typing import Optional
from tensorpc.core.datamodel.funcs import not_null
@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    type: str = ""
    is_custom_node: bool = False
    text: Optional[str] = None

@dataclasses.dataclass
class BasicFlowModelRoot(models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel]):
    pass 

class App:
    class Config(plus.AppInMemory.Config):
        is_horizontal = False
        height = "800px"
    @mark_create_layout
    def my_layout(self):
        items = [
            mui.MenuItem(id="input", label="Add Input Node"),
            mui.MenuItem(id="output", label="Add Output Node"),
            mui.MenuItem(id="default", label="Add Default Node"),
            mui.MenuItem(id="custom", label="Add Custom Node"),
        ]
        self.graph = flowui.Flow([], [], [
            flowui.MiniMap(),
            flowui.Controls(),
            flowui.Background()
        ]).prop(paneContextMenuItems=items)
        jv = mui.JsonViewer()
        jv.bind_fields(data="$")
        self.graph.event_pane_context_menu.on(self.add_node)
        self.dm = mui.DataModel(BasicFlowModelRoot({}, {}), [
            mui.VBox([
                self.graph,
            ]).prop(flex=1),
            mui.VBox([
                jv,
            ]).prop(flex=1, overflow="auto", fontSize="12px"),

        ])
        binder = models.flow.BaseFlowModelBinder(
            self.graph, self.dm.get_model, 
            self.dm.get_draft(), 
            self.model_to_ui_node)
        binder.bind_flow_comp_with_base_model(self.dm)
        return mui.HBox([
            self.dm,
        ]).prop(width="100%", height="100%", overflow="hidden")
    
    async def add_node(self, data: flowui.PaneContextMenuEvent):
        item_id = data.itemId
        node_type = item_id
        pos = data.flowPosition
        if pos is None:
            return 
        node_id = self.graph.create_unique_node_id(node_type)
        draft = self.dm.get_draft()
        if node_type != "custom":
            new_node = NodeModel(type=node_type, id=node_id, text=node_type, position=pos)
        else:
            new_node = NodeModel(type="app", id=node_id, position=pos, text="Editable Text Here", is_custom_node=True)
        draft.nodes[node_id] = new_node

    def model_to_ui_node(self, flow: models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel], node_id: str):
        node = flow.nodes[node_id]
        # we need to create component outside datamodel,
        # we dont support saving component to data model.
        if node.is_custom_node:
            draft = self.dm.get_draft()
            node_draft = draft.nodes[node.id]
            # we need to wrap optional value with defaults when use `cformat`
            node_type_draft = not_null(node_draft.type, "unknown")
            node_width_draft = not_null(node_draft.width, -1)
            node_height_draft = not_null(node_draft.height, -1)
            node_pos_x_draft = not_null(node_draft.position.x, -1)
            node_pos_y_draft = not_null(node_draft.position.y, -1)
            base_info_draft = create_literal_draft("### %s Node\n\n") % node_type_draft
            wh_draft = create_literal_draft("Width: `%d`, Height: `%d`") % (node_width_draft, node_height_draft)
            pos_draft = create_literal_draft("Position: `%d, %d`") % (node_pos_x_draft, node_pos_y_draft)

            comp = mui.VBox([
                flowui.NodeResizer().prop(minWidth=50, minHeight=50),
                mui.Markdown().bind_fields(value=base_info_draft),
                mui.Markdown().bind_fields(value=wh_draft),
                mui.Markdown().bind_fields(value=pos_draft),
                mui.Markdown().bind_fields(value=node_draft.text),
            ])
            ui_node = flowui.Node(type="app", id=node.id, data=flowui.NodeData(component=comp), position=node.position)
            return ui_node
        else:
            node_dict = dataclasses.asdict(node)
            node_dict.pop("text")
            return flowui.Node(data=flowui.NodeData(label=node.text), **node_dict)


```

You can find that app above can show node location and size, but size/location are debounced. This is expected because these events are updated in backend.