## Node Code Editor

`mui.MonacoEditor` is another uncontrolled component. Here we create markdown node, and attach a `mui.MonacoEditor` to it when the node is selected, and update node content when you save from editor.

### Model Design

* Node Model

```py
@dataclasses.dataclass
class NodeModel(flowui.BaseNodeModel):
    content: str = ""
```

* Code Editor

Since `mui.MonacoEditor` is uncontrolled component, we need to implement 
`Model -> Editor` and `Editor -> Model` data flow. We have provide a simple method to bind three props of editor: `value`, `language` and `path`.


### Example

```Python
from tensorpc.dock import mui, three, plus, appctx, mark_create_layout, flowui, models
from tensorpc.core import dataclass_dispatch as dataclasses
from tensorpc.core.datamodel.draft import create_literal_draft
import tensorpc.core.datamodel.funcs as D

from typing import Optional
@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    content: str = "## Empty"

@dataclasses.dataclass
class BasicFlowModelRoot(models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel]):
    selected_node: Optional[str] = None


class App:
    class Config(plus.AppInMemory.Config):
        is_horizontal = False
        height = "800px"
    @mark_create_layout
    def my_layout(self):
        self.graph = flowui.Flow([], [], [
            flowui.MiniMap(),
            flowui.Controls(),
            flowui.Background()
        ])
        editor = mui.MonacoEditor("## WTF", "markdown", "default").prop(flex=1, minHeight=0, minWidth=0)
        init_nodes = {
            "md1": NodeModel(id="md1", position=flowui.XYPosition(0, 0), content="## Click Me To Edit"),
            "md2": NodeModel(id="md2", position=flowui.XYPosition(0, 200), content="## Click Me To Edit"),
        }
        editor_ct = mui.MatchCase.binary_selection(True, mui.VBox([
            editor,
        ]).prop(flex=1, overflow="hidden"))
        self.dm = mui.DataModel(BasicFlowModelRoot(init_nodes, {}), [
            mui.VBox([
                self.graph,
            ]).prop(flex=1),
            editor_ct,
        ])
        draft = self.dm.get_draft()
        editor_ct.bind_fields(condition=f"{draft.selected_node} != null")
        binder = models.flow.BaseFlowModelBinder(
            self.graph, self.dm.get_model, 
            draft, 
            self.model_to_ui_node)
        binder.bind_flow_comp_with_base_model(self.dm, draft.selected_node)
        code_draft = draft.nodes[draft.selected_node].content
        path_draft = D.not_null(draft.selected_node, "default")
        editor.bind_draft_change_uncontrolled(code_draft, path_draft=path_draft)
        return mui.HBox([
            self.dm,
        ]).prop(width="100%", height="100%", overflow="hidden")
    
    def model_to_ui_node(self, flow: models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel], node_id: str):
        node = flow.nodes[node_id]
        draft = self.dm.get_draft()
        comp = mui.VBox([
            mui.Markdown("### Markdown Node"),
            mui.Markdown().bind_fields(value=draft.nodes[node.id].content).prop(emoji=True),
        ])
        ui_node = flowui.Node(type="app", 
            id=node.id, 
            data=flowui.NodeData(component=comp), 
            position=node.position)
        return ui_node

```

We can see that the logic of our app is very simple with MVC pattern and draft system.