## Nested Flow

### Model Design

Firstly we need a way to query a nested flow in frontend, so we provide a builtin jmespath function `getitem_path` to query a object from a list of keys.

```py

@dataclasses.dataclass
class BasicFlowModelRoot(BasicFlowModel):
    cur_path: list[str] = dataclasses.field(default_factory=list)

nested_obj_draft = D.getitem_path(root_draft, root_draft.cur_path)

```

Now we can define a node with a optional nested flow:

```py

@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    flow: Optional["BasicFlowModel"] = None

```

The complete model design is:

```py
@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    flow: Optional["BasicFlowModel"] = None

@dataclasses.dataclass
class BasicFlowModel(models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel]):
    selected_node: Optional[str] = None

@dataclasses.dataclass
class BasicFlowModelRoot(BasicFlowModel):
    # example: ['nodes', 'node_id_0', flow, 'nodes', 'node_id_1', 'flow']
    cur_path: list[str] = dataclasses.field(default_factory=list)

```

### Example

```Python
from tensorpc.dock import mui, three, plus, appctx, mark_create_layout, flowui, models
from tensorpc.core import dataclass_dispatch as dataclasses
from tensorpc.core.datamodel.draft import create_literal_draft
import tensorpc.core.datamodel.funcs as D
from functools import partial
from tensorpc.core.tree_id import UniqueTreeIdForTree

from typing import Optional, Any

@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    label: str = ""
    flow: Optional["BasicFlowModel"] = None


@dataclasses.dataclass
class BasicFlowModel(models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel]):
    selected_node: Optional[str] = None

@dataclasses.dataclass
class BasicFlowModelRoot(BasicFlowModel):
    # example: ['nodes', 'node_id_0', flow, 'nodes', 'node_id_1']
    cur_path: list[str] = dataclasses.field(default_factory=list)

    def get_uid_from_path(self):
        return UniqueTreeIdForTree.from_parts(self.cur_path).uid_encoded

class App:
    class Config(plus.AppInMemory.Config):
        is_horizontal = False
        height = "800px"

    @mark_create_layout
    def my_layout(self):
        node_cm_items = [
            mui.MenuItem(id="nested", label="Enter Nested"),
        ]
        items = [
            mui.MenuItem(id="plain", label="Add Plain Node"),
            mui.MenuItem(id="nested", label="Add Nested Flow Node"),
        ]

        self.graph = flowui.Flow([], [], [
            flowui.MiniMap(),
            flowui.Controls(),
            flowui.Background(),
        ]).prop(nodeContextMenuItems=node_cm_items, paneContextMenuItems=items)
        self.graph_preview = flowui.Flow([], [], [
            flowui.MiniMap(),
            flowui.Controls(),
            flowui.Background(),
        ]).prop(paneContextMenuItems=items)

        self.graph.event_node_context_menu.on(self.handle_node_cm)
        nested_node_1 = {
            "n3": NodeModel(id="n3", position=flowui.XYPosition(0, 0), label="Nested 1 node 1"),
            "n4": NodeModel(id="n4", position=flowui.XYPosition(0, 100), label="Nested 1 node 2"),
        }
        nested_node_0 = {
            "n1": NodeModel(id="n1", position=flowui.XYPosition(0, 0), label="Nested 0 node 1"),
            "n2": NodeModel(id="n2", position=flowui.XYPosition(0, 100), flow=BasicFlowModel(nested_node_1, {})),
        }
        init_nodes = {
            "n5": NodeModel(id="n5", position=flowui.XYPosition(0, 0), flow=BasicFlowModel(nested_node_0, {})),
        }
        path_breadcrumb = mui.Breadcrumbs([]).prop(keepHistoryPath=True)
        detail = mui.Typography("")
        detail_ct = mui.MatchCase.binary_selection(True, mui.VBox({
            "flow": mui.HBox([
                mui.Typography("Nested Flow Preview").prop(variant="h6"),
                mui.Button("Enter Flow"),
                detail,
            ]),
            "flow_preview": self.graph_preview,
        }).prop(flex=1, overflow="hidden"))

        self.dm = mui.DataModel(BasicFlowModelRoot(init_nodes, {}), [
            mui.VBox([
                mui.HBox([
                    path_breadcrumb
                ]).prop(minHeight="24px"),
                self.graph,
            ]).prop(flex=1),
            detail_ct,
        ])

        draft = self.dm.get_draft()
        cur_model_draft = D.getitem_path_dynamic(draft, draft.cur_path, Optional[BasicFlowModel])
        # determine if we should show preview, only show preview when some node is selected in main flow and it has a nested flow
        should_show_preview = D.logical_and(cur_model_draft.selected_node != None, cur_model_draft.nodes[cur_model_draft.selected_node].flow != None)
        detail_ct.bind_fields(condition=should_show_preview)
        # use main flow path and selected node to construct preview flow path
        prev_path_draft_if_exist = draft.cur_path + ["nodes"] + D.create_array(cur_model_draft.selected_node) + ["flow"]
        # also if selected node is None, we should use empty path
        prev_path_draft = D.where(cur_model_draft.selected_node == None, [], prev_path_draft_if_exist, return_type=list[str])
        detail.bind_fields(value=f"join('.', {prev_path_draft})")
        # finally we can get nested preview flow draft from preview path
        preview_model_draft = D.getitem_path_dynamic(draft, prev_path_draft, Optional[BasicFlowModel])
        self.graph.event_pane_context_menu.on(partial(self.add_node, target_flow_draft=cur_model_draft))
        self.graph_preview.event_pane_context_menu.on(partial(self.add_node, target_flow_draft=preview_model_draft))
        # draft only support raw path, so we use [1::3] to convert from raw path to real node path
        # we also need to add root to the beginning
        path_breadcrumb.bind_fields(value=f"concat(`[\"root\"]`, {draft.cur_path}[1::3])")
        path_breadcrumb.event_change.on(self.handle_breadcrumb_click)
        # since we may switch preview flow repeatedly, we need to set a unique flow id to avoid handle wrong frontend event
        # e.g. the size/position change event is debounced
        binder = models.flow.BaseFlowModelBinder(
            self.graph, self.dm.get_model, 
            cur_model_draft, 
            self.model_to_ui_node,
            flow_uid_getter=lambda: self.dm.get_model().get_uid_from_path())
        binder.bind_flow_comp_with_base_model(self.dm, cur_model_draft.selected_node)
        preview_binder = models.flow.BaseFlowModelBinder(
            self.graph_preview, self.dm.get_model, 
            preview_model_draft, 
            self.model_to_ui_node,
            flow_uid_getter=partial(self._get_preview_flow_uid, prev_path_draft))
        preview_binder.bind_flow_comp_with_base_model(self.dm, preview_model_draft.selected_node)

        return mui.HBox([
            self.dm,
        ]).prop(width="100%", height="100%", overflow="hidden")
    
    def _get_preview_flow_uid(self, path_draft):
        path = D.evaluate_draft(path_draft, self.dm.model)
        if path is None:
            return "root"
        return UniqueTreeIdForTree.from_parts(path).uid_encoded

    def model_to_ui_node(self, flow: BasicFlowModel, node_id: str):
        node = flow.nodes[node_id]
        draft = self.dm.get_draft()
        comp = mui.VBox([
            mui.Typography(f"Node-{node.label}" if node.flow is None else "Nested Flow"),
        ])
        ui_node = flowui.Node(type="app", 
            id=node.id, 
            data=flowui.NodeData(component=comp, label=node.label), 
            position=node.position)
        return ui_node

    async def handle_node_cm(self, data: flowui.NodeContextMenuEvent):
        item_id = data.itemId
        node_id = data.nodeId

        cur_path_val = self.dm.model.cur_path
        new_path_val = cur_path_val + ['nodes', node_id, 'flow']
        new_logic_path = new_path_val[1::3]
        # validate node contains nested flow
        cur_model = self.dm.model
        for item in new_logic_path:
            cur_model = cur_model.nodes[item].flow
            if cur_model is None:
                return

        draft = self.dm.get_draft()
        draft.cur_path = new_path_val

    def handle_breadcrumb_click(self, data: list[str]):
        logic_path = data[1:] # remove root
        res_path: list[str] = []
        for item in logic_path:
            res_path.extend(['nodes', item, 'flow'])
        draft = self.dm.get_draft()
        draft.cur_path = res_path

    def add_node(self, data: flowui.PaneContextMenuEvent, target_flow_draft: Any):
        item_id = data.itemId
        node_type = item_id
        pos = data.flowPosition
        print(f"Add Node: {node_type} at {pos}")
        if pos is None:
            return 
        node_id = self.graph.create_unique_node_id(node_type)
        draft = self.dm.get_draft()
        if node_type == "nested":
            new_node = NodeModel(id=node_id, position=pos, flow=BasicFlowModel({}, {}))
        else:
            new_node = NodeModel(id=node_id, position=pos, label="Plain")
        target_flow_draft.nodes[node_id] = new_node

```
