## Save Flow Model to Storage

We can use draft KV store to store node data in a custom storage. You can use `Annotated` with our metadata to specify the storage meta and provide store implementation to `mui.DataModel`.


> Note: You can safety override parent field with same type and `Annotated` to attach new metadata to a parent field.

We will use example from `Nested Flow` to demonstrate how to store flow model in a custom storage. For better visualization, we will replace preview of nested flow with store data json.

### Storage Design

Assume we use a file-like storage. When our graph becomes huge, read/change/write update will be slow, so we need to store node state to specific path (file). 

Fortunately, we have a builtin `DraftStoreMapMeta` to indicate draft store automatically handle this. You only need to implement a suitable store backend.


### Model Design

```py
@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    flow: Optional["BasicFlowModel"] = None

@dataclasses.dataclass
class BasicFlowModel(models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel]):
    nodes: Annotated[dict[str, NodeModel], DraftStoreMapMeta()]
    selected_node: Optional[str] = None

@dataclasses.dataclass
class BasicFlowModelRoot(BasicFlowModel):
    # example: ['nodes', 'node_id_0', flow, 'nodes', 'node_id_1', 'flow']
    cur_path: list[str] = dataclasses.field(default_factory=list)

```

### Example

```Python
from tensorpc.dock import mui, three, plus, appctx, mark_create_layout, flowui, models
from tensorpc.core import dataclass_dispatch as dataclasses
from tensorpc.core.datamodel.draft import create_literal_draft
import tensorpc.core.datamodel.funcs as D
from functools import partial
from tensorpc.core.tree_id import UniqueTreeIdForTree

from typing import Optional, Any, Annotated
from tensorpc.core.datamodel.draftstore import (DraftFileStorage,
                                                DraftFileStoreBackendInMemory,
                                                DraftStoreMapMeta,
                                                get_splitted_update_model_ops,
                                                DraftUpdateOp)

@dataclasses.dataclass
class NodeModel(models.flow.BaseNodeModel):
    flow: Optional["BasicFlowModel"] = None
    label: str = ""

# we need to use kw_only and same type to override parent field.
@dataclasses.dataclass(kw_only=True)
class BasicFlowModel(models.flow.BaseFlowModel[NodeModel, models.flow.BaseEdgeModel]):
    nodes: Annotated[dict[str, NodeModel], DraftStoreMapMeta(base64_key=False)]
    selected_node: Optional[str] = None

@dataclasses.dataclass(kw_only=True)
class BasicFlowModelRoot(BasicFlowModel):
    # example: ['nodes', 'node_id_0', flow, 'nodes', 'node_id_1']
    cur_path: list[str] = dataclasses.field(default_factory=list)

    def get_uid_from_path(self):
        return UniqueTreeIdForTree.from_parts(self.cur_path).uid_encoded

class DraftFileStoreBackendInMemoryForTutorial(DraftFileStoreBackendInMemory):
    """Show content to `mui.JsonViewer` when write, for debug"""
    def __init__(self, jv: mui.JsonViewer):
        super().__init__()
        self.jv = jv
        
    async def write(self, path: str, data: Any) -> None:
        await super().write(path, data)
        await self.jv.write(self._data)

    async def update(self, path: str, ops: list[DraftUpdateOp]) -> None:
        await super().update(path, ops)
        await self.jv.write(self._data)

    async def remove(self, path: str) -> None:
        await super().remove(path)
        await self.jv.write(self._data)


class App:
    class Config(plus.AppInMemory.Config):
        is_horizontal = False
        height = "800px"

    @mark_create_layout
    def my_layout(self):
        node_cm_items = [
            mui.MenuItem(id="nested", label="Enter Nested"),
        ]
        items = [
            mui.MenuItem(id="plain", label="Add Plain Node"),
            mui.MenuItem(id="nested", label="Add Nested Flow Node"),
        ]

        self.graph = flowui.Flow([], [], [
            flowui.MiniMap(),
            flowui.Controls(),
            flowui.Background(),
        ]).prop(nodeContextMenuItems=node_cm_items, paneContextMenuItems=items)
        self.graph_preview = flowui.Flow([], [], [
            flowui.MiniMap(),
            flowui.Controls(),
            flowui.Background(),
        ]).prop(paneContextMenuItems=items)

        self.graph.event_node_context_menu.on(self.handle_node_cm)
        nested_node_1 = {
            "n3": NodeModel(id="n3", position=flowui.XYPosition(0, 0), label="Nested 1 node 1"),
            "n4": NodeModel(id="n4", position=flowui.XYPosition(0, 100), label="Nested 1 node 2"),
        }
        nested_node_0 = {
            "n1": NodeModel(id="n1", position=flowui.XYPosition(0, 0), label="Nested 0 node 1"),
            "n2": NodeModel(id="n2", position=flowui.XYPosition(0, 100), flow=BasicFlowModel(nodes=nested_node_1, edges={})),
        }
        init_nodes = {
            "n5": NodeModel(id="n5", position=flowui.XYPosition(0, 0), flow=BasicFlowModel(nodes=nested_node_0, edges={})),
        }
        path_breadcrumb = mui.Breadcrumbs([]).prop(keepHistoryPath=True)
        jv = mui.JsonViewer()

        self.dm = mui.DataModel(BasicFlowModelRoot(nodes=init_nodes, edges={}), [
            mui.VBox([
                mui.HBox([
                    path_breadcrumb
                ]).prop(minHeight="24px"),
                self.graph,
            ]).prop(flex=1),
            mui.VBox([
                jv,
            ]).prop(flex=1, overflow="auto", fontSize="12px"),
        ])
        # we use a in-memory store to see stored data.
        # you can implement your own path-based store or use `AppDraftFileStoreBackend` to store data in app storage/global storage.
        # WARNING: currently our storage include `AppDraftFileStoreBackend` don't support coop.
        in_memory_store = DraftFileStoreBackendInMemoryForTutorial(jv)
        self.dm.connect_draft_store("test", in_memory_store)
        draft = self.dm.get_draft()
        cur_model_draft = D.getitem_path_dynamic(draft, draft.cur_path, Optional[BasicFlowModel])
        self.graph.event_pane_context_menu.on(partial(self.add_node, target_flow_draft=cur_model_draft))
        # draft only support raw path, so we use [1::3] to convert from raw path to real node path
        # we also need to add root to the beginning
        path_breadcrumb.bind_fields(value=f"concat(`[\"root\"]`, {draft.cur_path}[1::3])")
        path_breadcrumb.event_change.on(self.handle_breadcrumb_click)
        # since we may switch preview flow repeatedly, we need to set a unique flow id to avoid handle wrong frontend event
        # e.g. the size/position change event is debounced
        binder = models.flow.BaseFlowModelBinder(
            self.graph, self.dm.get_model, 
            cur_model_draft, 
            self.model_to_ui_node,
            flow_uid_getter=lambda: self.dm.get_model().get_uid_from_path())
        binder.bind_flow_comp_with_base_model(self.dm, cur_model_draft.selected_node)
        return mui.HBox([
            self.dm,
        ]).prop(width="100%", height="100%", overflow="hidden")
    
    def model_to_ui_node(self, flow: BasicFlowModel, node_id: str):
        node = flow.nodes[node_id]
        draft = self.dm.get_draft()
        comp = mui.VBox([
            mui.Typography(f"Node-{node.label}" if node.flow is None else "Nested Flow"),
        ])
        ui_node = flowui.Node(type="app", 
            id=node.id, 
            data=flowui.NodeData(component=comp), 
            position=node.position)
        return ui_node

    async def handle_node_cm(self, data: flowui.NodeContextMenuEvent):
        item_id = data.itemId
        node_id = data.nodeId

        cur_path_val = self.dm.model.cur_path
        new_path_val = cur_path_val + ['nodes', node_id, 'flow']
        new_logic_path = new_path_val[1::3]
        # validate node contains nested flow
        cur_model = self.dm.model
        for item in new_logic_path:
            cur_model = cur_model.nodes[item].flow
            if cur_model is None:
                return

        draft = self.dm.get_draft()
        draft.cur_path = new_path_val

    def handle_breadcrumb_click(self, data: list[str]):
        logic_path = data[1:] # remove root
        res_path: list[str] = []
        for item in logic_path:
            res_path.extend(['nodes', item, 'flow'])
        draft = self.dm.get_draft()
        draft.cur_path = res_path

    def add_node(self, data: flowui.PaneContextMenuEvent, target_flow_draft: Any):
        item_id = data.itemId
        node_type = item_id
        pos = data.flowPosition
        print(f"Add Node: {node_type} at {pos}")
        if pos is None:
            return 
        node_id = self.graph.create_unique_node_id(node_type)
        draft = self.dm.get_draft()
        if node_type == "nested":
            new_node = NodeModel(id=node_id, position=pos, flow=BasicFlowModel(nodes={}, edges={}))
        else:
            new_node = NodeModel(id=node_id, position=pos, label="Plain")
        target_flow_draft.nodes[node_id] = new_node

```
